!2345678911234567892123456789312345678941234567895123456789612345678971234567898
!        0         0         0         0         0         0         0
!                   
! Module: atthkgen_mod
!
! Authors:  Andy Read - XMM-SSC-MPE
!          
! Description: The task atthkgen produces a periodic attitude HK file for an entire observation.
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module atthkgen_mod

  use oal
  use dal
  use param  
  use errorHandling

  implicit none 

  private 

  public atthkgen

contains 

  !---------------------------------------------------------------------
  subroutine atthkgen
  !---------------------------------------------------------------------

  type (SpacecraftAttitudeType)              :: att
  type (OdfInfoType)                         :: odfinfos
  type (ProposalInfoType)                    :: proposalinfo
  type(DataSetT)                             :: outhk
  type(TableT)                               :: hktab
  type(ColumnT)                              :: ahfra, ahfdec, ahfpa, &
                                                omra, omdec, ompa, &
                                                dahfpnt, dompnt, dahfom, time

  character(len=10), parameter               :: CodeVersion = '1.17.3+x'
  character(len=*),  parameter               :: taskname = "atthkgen"
  character(len=256)                         :: history
  character(len=80)                          :: obsid !, expid

  integer				                                :: nsteps, perc_done, & !nn, 
                                                n, nahf, nom, nahfom, &
                                                nahftot, nomtot, nahfomtot

  real(kind=double)                          :: t, tbegin, tend, timestep, &
                                                tot_ahfra, &  !rapnt, decpnt, 
                                                tot_ahfdec, tot_ahfpa, &
                                                tot_omra, tot_omdec, &
                                                tot_ompa, tot_dahfpnt, &
                                                tot_dompnt, tot_dahfom, &
                                                av_ahfra, &
                                                av_ahfdec, av_ahfpa, &
                                                av_omra, av_omdec, &
                                                av_ompa, av_dahfpnt, &
                                                av_dompnt, av_dahfom, &
                                                md_ahfra, &
                                                md_ahfdec, md_ahfpa, &
                                                md_omra, md_omdec, &
                                                md_ompa, md_dahfpnt, &
                                                md_dompnt, md_dahfom
                                                
  real(kind=double), dimension(:), pointer   :: p_ahfra, p_ahfdec, p_ahfpa, &
                                                p_omra, p_omdec, p_ompa, &
                                                p_dahfpnt, p_dompnt, &
                                                p_dahfom, p_time

  real(kind=double), parameter               :: pi = 3.1415926535897932384d0, &
                                                radian = 180.d0/pi

  real(kind=double), dimension(:), allocatable :: xarr, yarr, zarr

!******************************************************************************
!	START 
!*************************
    if (verbosity(AppMsg) >= Noisy) &
       call message(AppMsg,Verbose,'CodeVersion: ATTHKGEN-'//trim(CodeVersion))
!--------------------------------------------------------
!	OPEN OAL, get start/end times and nominal RA/dec
!--------------------------------------------------------
    call OAL_odfInfo(odfinfos)
    write(errstr,*) "Obs. runs ",odfinfos%observationStartTime," to ",odfinfos%observationEndTime
    call message(AppMsg,Verbose)
    tbegin = toTimeTag(odfinfos%observationStartTime)
    tend = toTimeTag(odfinfos%observationEndTime)
    write(errstr,*) "Observation start time: ",tbegin
    call message(AppMsg,Verbose)
    write(errstr,*) "Observation end time: ",tend
    call message(AppMsg,Verbose)
    call OAL_proposalInfo(proposalinfo)

!---------------------------------------------------
!	CREATE HK FILE
!---------------------------------------------------
    t = tbegin
    timestep = realParameter("timestep")
    if (timestep <= 0) then 
      write (errstr,*) "Parameter timestep (",timestep,") should be greater than 0"
      call error("badTimestep")
    endif
    nsteps  = 1 + nint ((tend - tbegin)/timestep)
    write(errstr,*) "Increments in ",nsteps-1,"(+ 1) steps of ",timestep," seconds"
    call message(AppMsg,Verbose)

    outhk = dataSet(stringParameter("atthkset"),CREATE)
    hktab = addTable(outhk,"ATTHK",nsteps)
    time = addColumn(hktab,"TIME",Real64,"sec",comment="time")
    ahfra = addColumn(hktab,"AHFRA",Real64,"degrees",comment="AHF Right Ascension")
    ahfdec = addColumn(hktab,"AHFDEC",Real64,"degrees",comment="AHF Declination")
    ahfpa = addColumn(hktab,"AHFPA",Real64,"degrees",comment="AHF Position Angle")
    omra = addColumn(hktab,"OMRA",Real64,"degrees",comment="OM Right Ascension")
    omdec = addColumn(hktab,"OMDEC",Real64,"degrees",comment="OM Declination")
    ompa = addColumn(hktab,"OMPA",Real64,"degrees",comment="OM Position Angle")
    dahfpnt = addColumn(hktab,"DAHFPNT",Real64,"degrees",comment="Pointing Difference (AHF-PNT)")
    dompnt = addColumn(hktab,"DOMPNT",Real64,"degrees",comment="Pointing Difference (OM-PNT)")
    dahfom = addColumn(hktab,"DAHFOM",Real64,"degrees",comment="Pointing Difference (AHF-OM)")

    p_time => real64Data(time)
    p_ahfra => real64Data(ahfra)
    p_ahfdec => real64Data(ahfdec)
    p_ahfpa => real64Data(ahfpa)
    p_omra => real64Data(omra)
    p_omdec => real64Data(omdec)
    p_ompa => real64Data(ompa)
    p_dahfpnt => real64Data(dahfpnt)
    p_dompnt => real64Data(dompnt)
    p_dahfom => real64Data(dahfom)
!---------------------------------------------------
!	Check nulls
!---------------------------------------------------
    call nullcheck(nullable(ahfra))

!---------------------------------------------------
!	SET-UPS BEFORE MAIN LOOP (1)
!---------------------------------------------------
    n = 0
    nahf = 0
    nom = 0
    nahfom = 0
    nahftot = 0
    nomtot = 0
    nahfomtot = 0
    tot_ahfra = 0.0
    tot_ahfdec = 0.0
    tot_ahfpa = 0.0
    tot_omra = 0.0
    tot_omdec = 0.0
    tot_ompa = 0.0
    tot_dahfom = 0.0
    av_ahfra = 400.0
    av_ahfdec = 100.0
    av_ahfpa = 400.0
    av_omra = 400.0
    av_omdec = 100.0
    av_ompa = 400.0
    av_dahfom = 400.0
    md_ahfra = 400.0
    md_ahfdec = 100.0
    md_ahfpa = 400.0
    md_omra = 400.0
    md_omdec = 100.0
    md_ompa = 400.0
    md_dahfom = 400.0
    perc_done = 20
!---------------------------------------------------
!	MAIN AHF/OM LOOP (loop 1)
!---------------------------------------------------
    do n = 1, nsteps
      if (n >= perc_done * nsteps /100) then 
        if (perc_done < 100) then
          write(errstr,*) "",perc_done,"% completed of 1st run (AHF/OM)"
        else
          write(errstr,*) perc_done,"% completed of 1st run (AHF/OM)"
        endif
        call message(AppMsg,Sparse)
        perc_done = perc_done + 20
      endif
      p_time(n-1) = t
      !--------------------------------!
      !  Attitude and quality (AHF)    !
      !--------------------------------!
      call OAL_setState(attitudeFromAhf = .TRUE.)	! AHF
      if (.not. OAL_getAttitude(t, att)) then 		! attitude bad
        call setToNull(ahfra,n-1)
        call setToNull(ahfdec,n-1)
        call setToNull(ahfpa,n-1)
        call setToNull(dahfpnt,n-1)
      else						! attitude good
        p_ahfra(n-1) = att%ra * radian
        p_ahfdec(n-1) = att%dec * radian 
        p_ahfpa(n-1) = att%apos * radian
        !--------------------------------!
        !  Averages + Stats (AHF)        !
        !--------------------------------!
        nahf = nahf + 1
        tot_ahfra = tot_ahfra + p_ahfra(n-1)
        tot_ahfdec = tot_ahfdec + p_ahfdec(n-1)
        tot_ahfpa = tot_ahfpa + p_ahfpa(n-1)
      endif
      !--------------------------------!
      !  Attitude and quality (OM)     !
      !--------------------------------!
      call OAL_setState(attitudeFromAhf = .FALSE.)	! OM
      if (.not. OAL_getAttitude(t, att)) then 		! attitude bad
        call setToNull(omra,n-1)
        call setToNull(omdec,n-1)
        call setToNull(ompa,n-1)
        call setToNull(dompnt,n-1)
      else						! attitude good
        p_omra(n-1) = att%ra * radian
        p_omdec(n-1) = att%dec * radian
        p_ompa(n-1) = att%apos * radian
        !--------------------------------!
        !  Averages + Stats (OM)         !
        !--------------------------------!
        nom = nom + 1
        tot_omra = tot_omra + p_omra(n-1)
        tot_omdec = tot_omdec + p_omdec(n-1)
        tot_ompa = tot_ompa + p_ompa(n-1)
      endif
      !--------------------------------!
      !  Difference (AHF - OM)         !
      !--------------------------------!
      if (.not.isNull(dahfpnt,n-1) .and. .not.isNull(dompnt,n-1)) then 
        if ( ABS(p_ahfra(n-1)  - p_omra(n-1))  < 1.0e-06 .and. &
             ABS(p_ahfdec(n-1) - p_omdec(n-1)) < 1.0e-06) then 
          p_dahfom(n-1) = 0.0
        else
          p_dahfom(n-1) = (cos(p_ahfra(n-1)/radian)*cos(p_ahfdec(n-1)/radian)* &
                         cos(p_omra(n-1)/radian)*cos(p_omdec(n-1)/radian)) + &
                        (sin(p_ahfra(n-1)/radian)*cos(p_ahfdec(n-1)/radian)* &
                         sin(p_omra(n-1)/radian)*cos(p_omdec(n-1)/radian)) + &
                        (sin(p_ahfdec(n-1)/radian)*sin(p_omdec(n-1)/radian))
          if (p_dahfom(n-1) > +1.0) p_dahfom(n-1) = +1.0
          if (p_dahfom(n-1) < -1.0) p_dahfom(n-1) = -1.0
          p_dahfom(n-1) = radian * abs(acos(p_dahfom(n-1)))
        endif
        !--------------------------------!
        !  Averages + Stats (AHF - OM)   !
        !--------------------------------!
        nahfom = nahfom + 1
        tot_dahfom = tot_dahfom + p_dahfom(n-1)
      else
        call setToNull(dahfom,n-1)
      endif
      t = t + timestep
    enddo
    nahftot = nahf
    nomtot = nom
    nahfomtot = nahfom

! write out the abovementioned variables !!!!!

!---------------------------------------------------
!	END OF AHF/OM LOOP (loop 1)
!---------------------------------------------------

!---------------------------------------------------
!	Calculate AHF and OM averages
!---------------------------------------------------
    if (nahftot > 0) then 
      av_ahfra = tot_ahfra / nahftot
      av_ahfdec = tot_ahfdec / nahftot
      av_ahfpa = tot_ahfpa / nahftot
    else
      av_ahfra = 400.0
      av_ahfdec = 400.0
      av_ahfpa = 400.0
      av_dahfpnt = 400.0
      call warning("BadAhfAttitude","No good AHF attitude found")
    endif   
    if (nomtot > 0) then 
      av_omra = tot_omra / nomtot
      av_omdec = tot_omdec / nomtot
      av_ompa = tot_ompa / nomtot
    else
      av_omra = 400.0
      av_omdec = 400.0 
      av_ompa = 400.0 
      av_dompnt = 400.0 
      call warning("BadOmAttitude","No good OM attitude found")
    endif
    if (nahfomtot > 0) then 
      av_dahfom = tot_dahfom / nahfomtot
    else
      av_dahfom = 400.0
      call warning("BadAhfOmAttitude","No good simultaneous AHF and OM attitude found")
    endif   
!---------------------------------------------------
!	Calculate AHF and OM medians
!---------------------------------------------------
    allocate(xarr(1:nsteps),yarr(1:nsteps),zarr(1:nsteps))

    nahf = 0
    do n = 1, nsteps
      if (.not.isNull(ahfra,n-1) .and. .not.isNull(ahfdec,n-1) .and. &  
                                 .not.isNull(ahfpa,n-1)) then ! ahf att good
        nahf = nahf + 1
        xarr(nahf) = p_ahfra(n-1)
        yarr(nahf) = p_ahfdec(n-1)
        zarr(nahf) = p_ahfpa(n-1)
      endif
    enddo
    call median0(xarr,nahf,md_ahfra)
    call median0(yarr,nahf,md_ahfdec)
    call median0(zarr,nahf,md_ahfpa)

    nom = 0
    do n = 1, nsteps
      if (.not.isNull(omra,n-1) .and. .not.isNull(omdec,n-1) .and. &  
                                    .not.isNull(ompa,n-1)) then   ! om att good
        nom = nom + 1
        xarr(nom) = p_omra(n-1)
        yarr(nom) = p_omdec(n-1)
        zarr(nom) = p_ompa(n-1)
      endif
    enddo
    call median0(xarr,nom,md_omra)
    call median0(yarr,nom,md_omdec)
    call median0(zarr,nom,md_ompa)

    nahfom = 0
    do n = 1, nsteps
      if (.not.isNull(omra,n-1)  .and. .not.isNull(omdec,n-1)  .and. &  
          .not.isNull(ahfra,n-1) .and. .not.isNull(ahfdec,n-1) .and. &  
          .not.isNull(ahfpa,n-1) .and. .not.isNull(ompa,n-1))  then   ! om +ahf att good
        nahfom = nahfom + 1
        xarr(nahfom) = p_dahfom(n-1)
      endif
    enddo
    call median0(xarr,nahfom,md_dahfom)

    if (nahf < 1) then 
      md_ahfra = 400.0
      md_ahfdec = 400.0 
      md_ahfpa = 400.0 
      md_dahfpnt = 400.0 
    endif
    if (nom < 1) then 
      md_omra = 400.0 
      md_omdec = 400.0 
      md_ompa = 400.0 
      md_dompnt = 400.0 
    endif
    if (nahfom < 1) then 
      md_dahfom = 400.0
    endif
!---------------------------------------------------
!	END OF AHF/OM MEDIAN LOOP 
!---------------------------------------------------
! write (*,*) md_ahfra,md_ahfdec,md_ahfpa
! write (*,*) md_omra,md_omdec,md_ompa
! write (*,*) md_dahfom
!---------------------------------------------------
!	SET UPS BEFORE MAIN LOOP (2)
!---------------------------------------------------
    n = 0
    nahf = 0
    nom = 0
    nahfom = 0
    tot_dahfpnt = 0.0
    tot_dompnt = 0.0
    av_dahfpnt = 400.0
    av_dompnt = 400.0
    md_dahfpnt = 400.0
    md_dompnt = 400.0
    perc_done=20
    t = tbegin
!---------------------------------------------------
!	MAIN D...PNT LOOP (loop 2)
!---------------------------------------------------
    do n = 1, nsteps
      if (n >= perc_done * nsteps /100) then 
        if (perc_done < 100) then
          write(errstr,*) "",perc_done,"% completed of 2nd run (PNT)"
        else
          write(errstr,*) perc_done,"% completed of 2nd run (PNT)"
        endif
        call message(AppMsg,Sparse)
        perc_done = perc_done + 20
      endif
      p_time(n-1) = t
      if (.not. isNull(dahfpnt,n-1)) then 
        !--------------------------------!
        !  Difference (AHF - pnt)        !
        !--------------------------------!
        p_dahfpnt(n-1) = (cos(p_ahfra(n-1)/radian)*cos(p_ahfdec(n-1)/radian)*cos(md_ahfra/radian)* &
                       cos(md_ahfdec/radian)) + (sin(p_ahfra(n-1)/radian)*cos(p_ahfdec(n-1)/radian)* &
                       sin(md_ahfra/radian)*cos(md_ahfdec/radian)) + &
                       (sin(p_ahfdec(n-1)/radian)*sin(md_ahfdec/radian))
        if (p_dahfpnt(n-1) > 1.0) p_dahfpnt(n-1) = 1.0
        if (p_dahfpnt(n-1) < -1.0) p_dahfpnt(n-1) = -1.0
        p_dahfpnt(n-1) = radian * abs(acos(p_dahfpnt(n-1)))

        !--------------------------------!
        !  Averages + Stats (AHF)        !
        !--------------------------------!
        nahf = nahf + 1
        tot_dahfpnt = tot_dahfpnt + p_dahfpnt(n-1)
      endif
      if (.not. isNull(dompnt,n-1)) then 
        !--------------------------------!
        !  Difference (OM - pnt)         !
        !--------------------------------!
        p_dompnt(n-1) = (cos(p_omra(n-1)/radian)*cos(p_omdec(n-1)/radian)*cos(md_omra/radian)* &
                       cos(md_omdec/radian)) + (sin(p_omra(n-1)/radian)*cos(p_omdec(n-1)/radian)* &
                       sin(md_omra/radian)*cos(md_omdec/radian)) + &
                       (sin(p_omdec(n-1)/radian)*sin(md_omdec/radian))
        if (p_dompnt(n-1) > 1.0) p_dompnt(n-1) = 1.0
        if (p_dompnt(n-1) < -1.0) p_dompnt(n-1) = -1.0
        p_dompnt(n-1) = radian * abs(acos(p_dompnt(n-1)))
        !--------------------------------!
        !  Averages + Stats (OM)         !
        !--------------------------------!
        nom = nom + 1
        tot_dompnt = tot_dompnt + p_dompnt(n-1)
      endif
      t = t + timestep
    enddo
!---------------------------------------------------
!	END OF  LOOP (loop 2)
!---------------------------------------------------
!---------------------------------------------------
!	Calculate D...PNT averages
!---------------------------------------------------
    if (nahftot > 0) then 
      av_dahfpnt = tot_dahfpnt / nahftot
    endif   
    if (nomtot > 0) then 
      av_dompnt = tot_dompnt / nomtot
    endif
!---------------------------------------------------
!	Calculate D...PNT medians
!---------------------------------------------------
    nahf = 0
    nom = 0
    do n = 1, nsteps
      if (.not.isNull(ahfra,n-1) .and. .not.isNull(ahfdec,n-1) .and. &  
                                 .not.isNull(ahfpa,n-1)) then ! ahf att good
        nahf = nahf + 1
        xarr(nahf) = p_dahfpnt(n-1)
      endif
      if (.not.isNull(omra,n-1) .and. .not.isNull(omdec,n-1) .and. &  
                                    .not.isNull(ompa,n-1)) then   ! om att good
        nom = nom + 1
        yarr(nom) = p_dompnt(n-1)
      endif
    enddo
    call median0(xarr,nahf,md_dahfpnt)
    call median0(yarr,nom,md_dompnt)

! write (*,*) md_dahfpnt,md_dompnt

    deallocate(xarr,yarr,zarr)

!---------------------------------------------------
!	END OF D...PNT MEDIAN LOOP 
!---------------------------------------------------

!---------------------------------------------------
!	Write out (and calc D..PNT) averages
!---------------------------------------------------
    if (nahftot > 0 .or. nomtot > 0) then 
      write(errstr,*) "********** Mean Values : **********"
    endif
    call message(AppMsg,Verbose)
    if (nahftot > 0) then 
      av_dahfpnt = tot_dahfpnt / nahftot
      write(errstr,*) "AHF RA  (",nahftot," data points): ", av_ahfra
      call message(AppMsg,Verbose)
      write(errstr,*) "AHF Dec (",nahftot," data points): ", av_ahfdec
      call message(AppMsg,Verbose)
      write(errstr,*) "AHF PA  (",nahftot," data points): ", av_ahfpa
      call message(AppMsg,Verbose)
      write(errstr,*) "(AHF-PNT) pointing diff.(",nahftot," data points): ", av_dahfpnt
      call message(AppMsg,Verbose)
    endif
    if (nomtot > 0) then 
      av_dompnt = tot_dompnt / nomtot
      write(errstr,*) "OM RA   (",nomtot," data points): ", av_omra
      call message(AppMsg,Verbose)
      write(errstr,*) "OM Dec  (",nomtot," data points): ", av_omdec
      call message(AppMsg,Verbose)
      write(errstr,*) "OM PA   (",nomtot," data points): ", av_ompa
      call message(AppMsg,Verbose)
      write(errstr,*) "(OM-PNT) pointing diff. (",nomtot," data points): ", av_dompnt
      call message(AppMsg,Verbose)
    endif
    if (nahfomtot > 0) then 
      write(errstr,*) "(AHF-OM) pointing diff. (",nahfomtot," data points): ", av_dahfom
      call message(AppMsg,Verbose)
    endif
    write(errstr,*) "********** Median Values : **********"
    if (nahftot > 0) then 
      call message(AppMsg,Verbose)
      write(errstr,*) "AHF RA  : ", md_ahfra
      call message(AppMsg,Verbose)
      write(errstr,*) "AHF Dec : ", md_ahfdec
      call message(AppMsg,Verbose)
      write(errstr,*) "AHF PA  : ", md_ahfpa
      call message(AppMsg,Verbose)
      write(errstr,*) "(AHF-PNT) pointing diff. : ", md_dahfpnt
      call message(AppMsg,Verbose)
    endif
    if (nomtot > 0) then 
      write(errstr,*) "OM RA   : ", md_omra
      call message(AppMsg,Verbose)
      write(errstr,*) "OM Dec  : ", md_omdec
      call message(AppMsg,Verbose)
      write(errstr,*) "OM PA   : ", md_ompa
      call message(AppMsg,Verbose)
      write(errstr,*) "(OM-PNT) pointing diff.  : ", md_dompnt
      call message(AppMsg,Verbose)
    endif
    if (nahfomtot > 0) then 
      write(errstr,*) "(AHF-OM) pointing diff.  : ", md_dahfom
      call message(AppMsg,Verbose)
    endif
!-------------------------------------------------
!	ADD ATTRIBUTES TO PRIMARY HEADER
!-------------------------------------------------
    if (verbosity(AppMsg) > Noisy) &
       call message(AppMsg,Verbose,'... writing PRIMARY keywords')
    call setAttribute(outhk,"TELESCOP","XMM","Telescope (mission) name")
!   call setAttribute(outhk,"INSTRUME","EPN","Instrument name")
    call setAttribute(outhk,"DETNAM"," ","PRIME or REDUNDANT")

!   if (OAL_hasIPPV("FILTER_WHEEL_POSITION")) then
!     filter = OAL_getIPPVstring("FILTER_WHEEL_POSITION")
!   else
!     call warning("BadFilter","No filter found")
!     filter = 'UNKNOWN'
!   endif
!   call setAttribute(outhk,"FILTER",filter,"Filter ID")

    if (verbosity(AppMsg) > Noisy) &
       call message(AppMsg,Verbose,'... writing PRIMARY keyword: OBS_MODE')
    if (odfInfos%isSlew) then
      call setAttribute(outhk,"OBS_MODE","SLEW", &
                     "Observation mode (pointing or slew)")
    else
      call setAttribute(outhk,"OBS_MODE","POINTING", &
                     "Observation mode (pointing or slew)") 
    endif

    if (verbosity(AppMsg) > Noisy) &
       call message(AppMsg,Verbose,'... writing PRIMARY keyword: OBS_ID')
    write(obsid,"(2a)") trim(odfinfos%proposalId),trim(odfinfos%observationId)
    call setAttribute(outhk,"OBS_ID",obsid,"Observation identifier")
!   write(expid,"(i3)") expinfos%number
!   call setAttribute(outfile,"EXP_ID",expid,"Exposure identifier")
!   if (inmode == 0 .or. inmode == 1 .or. inmode == 8) then
!     select case (inmode)
!     case(0)
!       call setAttribute(outhk,"CONTENT","EPIC IMAGING MODE EVENT LIST", &
!                       "Content of file")
!       call setAttribute(outhk,"DATAMODE","IMAGING", &
!                       "Instrument mode (IMAGING, TIMING, FAST, etc.)")
!     case(1)
!       call setAttribute(outhk,"CONTENT","EPIC TIMING MODE EVENT LIST", &
!                       "Content of file")
!       call setAttribute(outhk,"DATAMODE","TIMING", &
!                       "Instrument mode (IMAGING, TIMING, FAST, etc.)")
!     case(8)
!       call setAttribute(outhk,"CONTENT","EPIC BURST MODE EVENT LIST", &
!                       "Content of file")
!       call setAttribute(outhk,"DATAMODE","BURST", &
!                       "Instrument mode (IMAGING, TIMING, FAST, etc.)")
!     end select
!   endif
    call setAttribute(outhk,"ORIGIN","Leicester/SSC","origin of FITS file")
!   call setAttribute(outhk,"OBJECT",trim(proposalinfo%targetName), &
!                           "Name of observed object")
!   call setAttribute(outhk,"EXPOSURE", &
!                            toTimeTag(expInfos%actualEndTime(1:23)) - &
!	                           toTimeTag(expInfos%actualStartTime(1:23)),"s", &
!                           "Exposure")
    call setAttribute(outhk,"DATE-OBS",trim(odfinfos%observationStartTime), &
                            "Date observations were made")
    call setAttribute(outhk,"DATE-END",trim(odfinfos%observationEndTime), &
                            "Date observations ended")
!   call setAttribute(outhk,"OBSERVER",trim(proposalinfo%GoName), &
!                           "Name of PI")

    call setAttribute(outhk,"RA_OBJ",proposalinfo%targetRa,"deg", &
                            "Ra of the observed object")
    call setAttribute(outhk,"DEC_OBJ",proposalinfo%targetDec,"deg", &
                            "Dec of the observed object")
    call setAttribute(outhk,"RA_NOM",proposalinfo%boresightRa,"deg", &
                            "Ra of the observed object")

    call setAttribute(outhk,"DEC_NOM",proposalinfo%boresightDec,"deg", &
                            "Dec of the observed object")

!-------------------------------------------------
!	ADD COMMON ATTRIBUTES TO PRIMARY HEADER
!-------------------------------------------------
    if (verbosity(AppMsg) >= Noisy) &
       call message(AppMsg,Verbose,'... writing common PRIMARY keywords via OAL:')
    call OAL_addCommonAttributes(outhk)
    if (verbosity(AppMsg) >= Noisy) &
       call message(AppMsg,Verbose,'... written common PRIMARY keywords via OAL.')

!-------------------------------------------------
!	ADD AVERAGE ATTRIBUTES TO PRIMARY HEADER
!-------------------------------------------------
    if (verbosity(AppMsg) > Noisy) &
       call message(AppMsg,Verbose,'... writing average PRIMARY keywords')
    call setAttribute(outhk,"AAHFRA",av_ahfra,"deg", "Mean AHF RA")
    call setAttribute(outhk,"AAHFDEC",av_ahfdec,"deg", "Mean AHF Dec")
    call setAttribute(outhk,"AAHFPA",av_ahfpa,"deg", "Mean AHF PA")
    call setAttribute(outhk,"AOMRA",av_omra,"deg", "Mean OM RA")
    call setAttribute(outhk,"AOMDEC",av_omdec,"deg", "Mean OM Dec")
    call setAttribute(outhk,"AOMPA",av_ompa,"deg", "Mean OM PA")
    call setAttribute(outhk,"ADAHFPNT",av_dahfpnt,"deg", "Mean (AHF-PNT) pointing difference")
    call setAttribute(outhk,"ADOMPNT",av_dompnt,"deg", "Mean (OM-PNT) pointing difference")
    call setAttribute(outhk,"ADAHFOM",av_dahfom,"deg", "Mean (AHF-OM) pointing difference")

    call setAttribute(outhk,"MAHFRA",md_ahfra,"deg", "Median AHF RA")
    call setAttribute(outhk,"MAHFDEC",md_ahfdec,"deg", "Median AHF Dec")
    call setAttribute(outhk,"MAHFPA",md_ahfpa,"deg", "Median AHF PA")
    call setAttribute(outhk,"MOMRA",md_omra,"deg", "Median OM RA")
    call setAttribute(outhk,"MOMDEC",md_omdec,"deg", "Median OM Dec")
    call setAttribute(outhk,"MOMPA",md_ompa,"deg", "Median OM PA")
    call setAttribute(outhk,"MDAHFPNT",md_dahfpnt,"deg", "Median (AHF-PNT) pointing difference")
    call setAttribute(outhk,"MDOMPNT",md_dompnt,"deg", "Median (OM-PNT) pointing difference")
    call setAttribute(outhk,"MDAHFOM",md_dahfom,"deg", "Median (AHF-OM) pointing difference")

    call setAttribute(outhk,"NATT",nsteps,"", "Number of attitude entries")
    call setAttribute(outhk,"NGAHF",nahf,"", "Number of good AHF attitude entries")
    call setAttribute(outhk,"NGOM",nom,"", "Number of good OM attitude entries")
    call setAttribute(outhk,"NGAHFOM",nahfom,"", "Number of good AHF+OM attitude entries")
!-------------------------------------------------
!       ADD HISTORY (can put more information in)
!-------------------------------------------------
    write (history,'(a,i6)') "Number of attitude entries: ",nsteps
    call addHistory(outhk,history)
    write (history,'(a,i6)') "Number of good AHF attitude entries: ",nahf
    call addHistory(outhk,history)
    write (history,'(a,i6)') "Number of good OM attitude entries: ",nom
    call addHistory(outhk,history)
    write (history,'(a,i6)') "Number of good AHF+OM attitude entries: ",nahfom
    call addHistory(outhk,history)
    write (history,'(a,f7.4)') "Atthkgen timestep (seconds): ",timestep
    call addHistory(outhk,history)
!--------------------------------------------------
!	RELEASE FILE
!--------------------------------------------------
!    call release(time)
!    call release(ahfra)
!    call release(ahfdec)
!    call release(ahfpa)
!    call release(omra)
!    call release(omdec)
!    call release(ompa)
!    call release(dahfpnt)
!    call release(dompnt)
!    call release(dahfom)
!write (*,*) "here"
    call release(outhk)
!write (*,*) "here"
!--------------------------------------------------
!	*** END ***
    end subroutine atthkgen
!--------------------------------------------------

  !---------------------------------------------------------------------
    subroutine nullcheck (thisNullable)

    use dal
    type(NullableT),intent(in)  :: thisNullable

    write(errstr,*)   "Null defined?: ",nullDefined(thisNullable)
    call message(AppMsg,Noisy)

    end subroutine nullcheck
  !---------------------------------------------------------------------


SUBROUTINE median0(x, n, xmed)

! Find the median of X(1), ... , X(N), using as much of the quicksort
! algorithm as is needed to isolate it.
! N.B. On exit, the array X is partially ordered.
!
!     Latest revision - 26 November 1996 [alan miller]
!     minor adaption for XMMSAS 24-05-02 [mjf]
!
IMPLICIT NONE

INTEGER, INTENT(IN)                            :: n
REAL(kind=double), INTENT(INOUT), DIMENSION(:) :: x(1:n)
REAL(kind=double), INTENT(OUT)                 :: xmed

! Local variables

REAL(kind=double)    :: temp, xhi, xlo, xmax, xmin
LOGICAL              :: odd
INTEGER              :: hi, lo, nby2, nby2p1, mid, i, j, k

nby2 = n / 2
nby2p1 = nby2 + 1
odd = .true.
xmed = 0.0

!     HI & LO are position limits encompassing the median.

IF (n == 2 * nby2) odd = .false.
lo = 1
hi = n
IF (n < 3) THEN
  IF (n < 1) THEN
    xmed = 0.0
    RETURN
  END IF
  xmed = x(1)
  IF (n == 1) RETURN
  xmed = 0.5*(xmed + x(2))
  RETURN
END IF

!     Find median of 1st, middle & last values.

10 mid = (lo + hi)/2
xmed = x(mid)
xlo = x(lo)
xhi = x(hi)
IF (xhi < xlo) THEN          ! Swap xhi & xlo
  temp = xhi
  xhi = xlo
  xlo = temp
END IF
IF (xmed > xhi) THEN
  xmed = xhi
ELSE IF (xmed < xlo) THEN
  xmed = xlo
END IF

! The basic quicksort algorithm to move all values <= the sort key (XMED)
! to the left-hand end, and all higher values to the other end.

i = lo
j = hi
50 DO
  IF (x(i) >= xmed) EXIT
  i = i + 1
END DO
DO
  IF (x(j) <= xmed) EXIT
  j = j - 1
END DO
IF (i < j) THEN
  temp = x(i)
  x(i) = x(j)
  x(j) = temp
  i = i + 1
  j = j - 1

!     Decide which half the median is in.

  IF (i <= j) GO TO 50
END IF

IF (.NOT. odd) THEN
  IF (j == nby2 .AND. i == nby2p1) GO TO 130
  IF (j < nby2) lo = i
  IF (i > nby2p1) hi = j
  IF (i /= j) GO TO 100
  IF (i == nby2) lo = nby2
  IF (j == nby2p1) hi = nby2p1
ELSE
  IF (j < nby2p1) lo = i
  IF (i > nby2p1) hi = j
  IF (i /= j) GO TO 100

! Test whether median has been isolated.

  IF (i == nby2p1) RETURN
END IF
100 IF (lo < hi - 1) GO TO 10

IF (.NOT. odd) THEN
  xmed = 0.5*(x(nby2) + x(nby2p1))
  RETURN
END IF
temp = x(lo)
IF (temp > x(hi)) THEN
  x(lo) = x(hi)
  x(hi) = temp
END IF
xmed = x(nby2p1)
RETURN

! Special case, N even, J = N/2 & I = J + 1, so the median is
! between the two halves of the series.   Find max. of the first
! half & min. of the second half, then average.

130 xmax = x(1)
DO k = lo, j
  xmax = MAX(xmax, x(k))
END DO
xmin = x(n)
DO k = i, hi
  xmin = MIN(xmin, x(k))
END DO
xmed = 0.5*(xmin + xmax)

RETURN
END SUBROUTINE median0


end module atthkgen_mod
